"""
Compute average reward per action.

Author: Ghiyath Alaswad (based on a code by Nele Albers, https://doi.org/10.4121/22153898.v1)
Date: August 2024
"""

import numpy as np
import reward_utils as util


def compute_avg_reward(data, weights, cost, num_act):
    """
    Computes the average reward per action across all samples.

    Args:
        data (list): List of samples in the form <s0, s1, a, r>.
        weights (list): Weights applied to each reward component.
        cost (float): The cost factor associated with actions.
        num_act (int): Number of possible actions.

    Returns:
        np.array: Array of average rewards for each action.
    """
    num_samples = len(data)

    # Extract rewards and actions from the data
    list_of_rewards = list(np.array(data, dtype=object)[:, 3])
    list_of_actions = list(np.array(data, dtype=object)[:, 2])

    # Compute weighted rewards, considering the cost of actions
    reward_list = util.get_weighted_rewards_list(list_of_rewards, weights, list_of_actions, cost)

    # Update data with the computed rewards
    for i in range(len(reward_list)):
        data[i][3] = reward_list[i]

    rewards = np.zeros(num_act)
    trials = np.zeros(num_act)

    # Aggregate rewards and counts for each action
    for data_index in range(num_samples):
        action = data[data_index][2]
        rewards[action] += reward_list[data_index]
        trials[action] += 1

    # Calculate average reward per action, handling cases where there are no trials
    avg_reward = np.divide(rewards, trials, out=np.zeros_like(rewards), where=trials != 0)

    return avg_reward
