"""
Q-value computation for RL model.

Author: Ghiyath ALaswad (based on code by Nele Albers, https://doi.org/10.4121/22153898.v1)
Date: August 2024
"""
from copy import deepcopy
import itertools
import numpy as np
from scipy.sparse import lil_matrix, csr_matrix


def policy_evaluation(observation_space_size, discount_factor, trans_func,
                      reward_func, policy, q_vals,
                      update_tolerance=0.000000001):
    """
    Returns the q-values for a specific policy.

    Args:
        observation_space_size: number of observations
        discount_factor: discount factor of MDP
        trans_func: transition function
        reward_func: reward function
        policy: policy to compute q-values for
        update_tolerance: precision
    """
    q_vals_new = deepcopy(q_vals)
    num_act = len(trans_func[0])

    update = 1  # max. change in Q-value in one iteration
    while update > update_tolerance:
        update = 0
        for s in range(observation_space_size):
            for a in range(num_act):
                q_vals_new[s, a] = sum([trans_func[s, a, s_prime] * (
                        discount_factor * q_vals[s_prime, int(policy[s_prime])] + reward_func[s, a, s_prime]) for
                                        s_prime in range(observation_space_size)])

                update = max(update, abs(q_vals[s, a] - q_vals_new[s, a]))
        q_vals = deepcopy(q_vals_new)

    return q_vals_new


def get_Q_values_opt_policy(discount_factor, trans_func, reward_func):
    """
    Returns the Q-values for each state under the optimal policy.
    Args:
        discount_factor: discount factor of MDP
        trans_func: transition function (dim.: num_states x num_actions x num_states)
        reward_func: reward function (dim.: num_states x num_actions)
    """
    min_iterations = 100
    num_states = len(trans_func)
    num_act = len(trans_func[0])
    q_vals = np.zeros((num_states, num_act))

    policy = np.zeros(num_states)
    policy_new = np.ones(num_states)
    it = 0

    while (not np.array_equal(policy, policy_new)) or it < min_iterations:
        q_vals = policy_evaluation(num_states, discount_factor,
                                   trans_func, reward_func, policy, q_vals)
        policy = policy_new
        policy_new = np.array([np.argmax(q_vals[s]) for s in range(num_states)])
        it += 1

    return q_vals, policy_new


def compute_q_vals_dynamics(data, weights, cost, num_feat_values, num_act=2, discount_factor=0.85):
    """
    Compute Q-values and dynamics for reinforcement learning using value iteration with sparse matrices.

    Args:
        data: List of samples, where each sample is of the form [s0, s1, a, r].
        weights: Weights for the different reward components (not used here since reward is pre-computed).
        cost: The cost factor associated with actions (not used since reward is pre-computed).
        num_feat_values: List specifying the number of possible values for each feature.
        num_act: Number of possible actions.
        discount_factor: Discount factor for future rewards.

    Returns:
        q_values: Computed Q-values.
        reward_func: Reward function after weighting.
        trans_func: Transition function representing the dynamics.
        trans_counts: Count of transitions.
    """
    print("Data sample:", data[:5])  # Print the first few samples for debugging
    print("Expected number of features:", len(num_feat_values))
    # Ensure the data points are correctly structured
    for i, data_point in enumerate(data):
        if len(data_point[0]) != len(num_feat_values) or len(data_point[1]) != len(num_feat_values):
            print(f"Data point {i} is incorrectly structured: {data_point}")

    # Determine the minimum and maximum values for each feature across all states
    min_values = [min([min(data_point[0][i], data_point[1][i]) for data_point in data]) for i in
                  range(len(num_feat_values))]
    max_values = [max([max(data_point[0][i], data_point[1][i]) for data_point in data]) for i in
                  range(len(num_feat_values))]

    # Convert min_values and max_values to integers
    min_values = [int(np.floor(value)) for value in min_values]
    max_values = [int(np.ceil(value)) for value in max_values]

    # Create a list of all possible states
    abstract_states = [list(i) for i in itertools.product(
        *[range(min_values[i], max_values[i] + 1) for i in range(len(num_feat_values))])]

    num_states = len(abstract_states)

    # Initialize sparse matrices in lil_matrix format for efficient construction
    trans_func = lil_matrix((num_states, num_act * num_states), dtype=np.float64)
    reward_func = lil_matrix((num_states, num_act * num_states), dtype=np.float64)
    reward_func_count = lil_matrix((num_states, num_act * num_states), dtype=np.float64)

    # Populate the transition and reward functions
    for data_point in data:
        current_state = data_point[0]
        next_state = data_point[1]
        action = data_point[2]
        reward = data_point[3]

        # Find the indices of the current and next states in the abstract_states list
        s_current_ind = abstract_states.index(current_state)
        s_next_ind = abstract_states.index(next_state)

        idx = action * num_states + s_next_ind

        # Update transition and reward functions
        trans_func[s_current_ind, idx] += 1
        reward_func[s_current_ind, idx] += reward
        reward_func_count[s_current_ind, idx] += 1

    print("Finished populating transition and reward functions.")

    # Convert to csr_matrix format for efficient operations
    trans_func = trans_func.tocsr()
    reward_func = reward_func.tocsr()
    reward_func_count = reward_func_count.tocsr()

    # Normalize the transition and reward functions
    for s_ind in range(num_states):
        for a in range(num_act):
            idx = slice(a * num_states, (a + 1) * num_states)
            if trans_func[s_ind, idx].sum() > 0:
                trans_func[s_ind, idx] = csr_matrix(trans_func[s_ind, idx] / trans_func[s_ind, idx].sum())

            # Convert sparse slice to dense format, perform division, and convert back to sparse
            reward_func_dense = reward_func[s_ind, idx].toarray()
            reward_func_count_dense = reward_func_count[s_ind, idx].toarray()

            # Avoid division by zero
            nonzero_mask = reward_func_count_dense > 0
            reward_func_dense[nonzero_mask] /= reward_func_count_dense[nonzero_mask]

            reward_func[s_ind, idx] = csr_matrix(reward_func_dense)

        if s_ind % 1000 == 0:
            print(f"Normalized {s_ind}/{num_states} states")

    print("Finished normalizing transition and reward functions.")

    # Value Iteration to compute Q-values
    q_values = np.zeros((num_states, num_act))
    max_iterations = 1000
    tolerance = 1e-6

    for iteration in range(max_iterations):
        q_values_old = q_values.copy()
        for s_ind in range(num_states):
            for a in range(num_act):
                idx = slice(a * num_states, (a + 1) * num_states)
                q_values[s_ind, a] = trans_func[s_ind, idx].dot(
                    reward_func[s_ind, idx].toarray().flatten() + discount_factor * np.max(q_values_old, axis=1))

            if s_ind % 1000 == 0:
                print(f"Value iteration {iteration + 1}: Updated {s_ind}/{num_states} states")

        # Check for convergence
        max_delta = np.max(np.abs(q_values - q_values_old))
        print(f"Iteration {iteration + 1}: max Q-value change = {max_delta:.6f}")

        if max_delta < tolerance:
            print(f"Convergence reached after {iteration + 1} iterations.")
            break

    return q_values, reward_func, trans_func, reward_func_count
