import pandas as pd
import numpy as np

FILE_PATH = 'preprocessed_data_with_reward_option1.csv'  # Path to the preprocessed data file
# FILE_PATH = 'preprocessed_data_with_reward_option2.csv'
STATE_NAMES = ['state_humansupport', 'state_importance', 'state_selfefficacy', 'state_energy']


def get_state_list(data):
    """
    Extracts state values for each session and returns them as a list.

    Args:
        data (DataFrame): Data containing the state information.

    Returns:
        list: A list of state representations for each session.
    """
    full_states = []
    for _, row in data.iterrows():
        full_state = [row[s] for s in STATE_NAMES]  # Extract values for each state variable
        full_states.append(full_state)
    return full_states


def get_next_states_rewards(data):
    """
    Computes the next states and corresponding rewards for each user.

    Args:
        data (DataFrame): The input data containing the states and initial rewards.

    Returns:
        list: A list of next state representations.
        list: A list of calculated rewards with sparse reward logic applied.
    """
    users = data.index.get_level_values('rand_id').unique()  # Identify unique users
    next_full_states, rewards = [], []
    for user in users:
        user_data = data.loc[user]
        user_states = user_data['s0'].tolist()  # Extract states for the user

        for i in range(len(user_states) - 1):
            next_full_states.append(user_states[i + 1])
            rewards.append(0)  # All intermediate steps get a reward of zero

        # Last state gets the actual reward
        terminal_state = [0] * len(STATE_NAMES)  # Replace NaN with a zero vector
        next_full_states.append(terminal_state)
        rewards.append(user_data['Weighted_Reward'].iloc[-1])  # Final reward

    return next_full_states, rewards


if __name__ == "__main__":
    # Load the preprocessed data
    df = pd.read_csv(FILE_PATH)

    # Define relevant columns, separating ID columns from numeric columns
    id_columns = ['rand_id', 'session_num']
    state_columns = ['state_humansupport', 'state_importance', 'state_selfefficacy', 'state_energy']
    action_column = ['human_support_after_session']
    reward_columns = ['Frequency_Change_Reward', 'Identity_Change_Reward', 'Weighted_Reward']

    # Only convert relevant columns to float
    df[state_columns + action_column + reward_columns] = df[state_columns + action_column + reward_columns].astype(
        float)

    df['s0'] = get_state_list(df)  # Generate a list of states for each session
    df = df[~df['s0'].apply(lambda x: any(np.isnan(x)))]  # Remove sessions with missing state values

    # Save all state data to a CSV file for further analysis
    df.to_csv("all_states.csv", index=False)

    # Compute the next state and corresponding reward for each session
    df['s1'], df['r'] = get_next_states_rewards(df.set_index('rand_id'))

    # Remove rows with missing reward values (cleaning up the dataset)
    df.dropna(inplace=True, subset=['r'])

    # Retain rand_id, session_num, state, action, and reward columns, and rename them accordingly
    df = df[id_columns + ['s0', 's1', 'human_support_after_session', 'r']]
    df.columns = ['rand_id', 'session_num', 's0', 's1', 'a', 'r']
    df['a'] = df['a'].astype(int)  # Ensure the action column is an integer

    # Save the final RL samples data to a CSV file
    df.to_csv("rl_samples.csv", index=False)

    print("RL samples saved to rl_samples.csv")

