"""
Feature selection based on the G-algorithm.

Author: Ghiyath Alaswad (based on code by Nele Albers, https://doi.org/10.4121/22153898.v1)
Date: August 2024
"""
from copy import deepcopy
import itertools
import math
import numpy as np
import random

import pandas as pd
from scipy import stats

import reward_utils as util
import Compute_RL_QValues as qval


def approx_dynamics_blocks(data, num_blocks_plus_1, num_act, blocks, blocks_plus_1, f, feat_sel):
    """Approximate reward and transition functions for blocks.

    Args:
        data (list): List with samples of the form <s0, s1, a, r>.
        num_blocks_plus_1 (int): Number of blocks when using 1 more than the so far selected features.
        num_act (int): Number of actions.
        blocks (list): Abstract states based on so far selected features.
        blocks_plus_1 (list): Abstract states based on selecting one more feature.
        f (int): Candidate feature.
        feat_sel (list): Already selected features.

    Returns:
        np-array: Weighted reward function.
        np-array: Transition function.
    """

    num_samples = len(data)

    trans_func = np.zeros((num_blocks_plus_1, num_act, num_blocks_plus_1))
    weighted_reward_func = np.zeros((num_blocks_plus_1, num_act, num_blocks_plus_1))
    reward_func_count = np.zeros((num_blocks_plus_1, num_act, num_blocks_plus_1))

    for s_ind, s in enumerate(blocks):  # For each block based on so far selected features

        for data_index in range(num_samples):  # For each data sample
            s_b = np.take(data[data_index][0], feat_sel)  # Values for already selected features
            s_prime_b = np.take(data[data_index][1], feat_sel)  # Next state values for already selected features
            a = data[data_index][2]  # Action
            r = data[data_index][3]  # Weighted reward

            # Construct the full state combinations as tuples
            s_combined = tuple(list(s_b) + [s])
            s_prime_combined = tuple(list(s_prime_b) + [data[data_index][1][f]])

            # Debugging: Check if the combined states exist in blocks_plus_1
            if s_combined not in blocks_plus_1:
                print(f"Error: Combined state {s_combined} not found in blocks_plus_1.")
                continue  # Skip this iteration but continue processing

            if s_prime_combined not in blocks_plus_1:
                print(f"Error: Combined state {s_prime_combined} not found in blocks_plus_1.")
                continue  # Skip this iteration but continue processing

            # Update statistics for transitions and rewards
            trans_func[blocks_plus_1.index(s_combined), a, blocks_plus_1.index(s_prime_combined)] += 1
            weighted_reward_func[blocks_plus_1.index(s_combined), a, blocks_plus_1.index(s_prime_combined)] += r
            reward_func_count[blocks_plus_1.index(s_combined), a, blocks_plus_1.index(s_prime_combined)] += 1

    # Normalize the reward and transition functions
    for b_ind in range(num_blocks_plus_1):
        for a in range(num_act):  # For each action
            summed = sum(trans_func[b_ind, a])
            if summed > 0:
                trans_func[b_ind, a] /= summed
            else:
                trans_func[b_ind, a] = np.ones(num_blocks_plus_1) / num_blocks_plus_1
            for s_prime_ind in range(num_blocks_plus_1):
                if reward_func_count[b_ind, a, s_prime_ind] > 0:
                    weighted_reward_func[b_ind, a, s_prime_ind] /= reward_func_count[b_ind, a, s_prime_ind]

    return weighted_reward_func, trans_func


def approx_dynamics_single_feature(data, num_values, num_act, f):
    """
    Approximate reward and transition functions for a single feature, handling sparse rewards.

    Args:
        data (list): List with samples of the form <s0, s1, a, r>.
        num_values (int): Number of possible values for the feature.
        num_act (int): Number of actions.
        f (int): Feature index to approximate dynamics for.

    Returns:
        np-array: reward function
        np-array: transition function
    """
    num_samples = len(data)

    trans_func = np.zeros((num_values, num_act, num_values))
    reward_func = np.zeros((num_values, num_act, num_values))
    reward_func_count = np.zeros((num_values, num_act, num_values))

    for s in range(num_values):  # For each possible value of the feature
        for data_index in range(num_samples):  # For each data sample
            if data[data_index][0][f] == s:
                a = data[data_index][2]
                s_prime = data[data_index][1][f]  # Feature value of next state
                r = data[data_index][3]

                # Update statistics
                trans_func[s, a, s_prime] += 1
                if isinstance(r, (list, np.ndarray)):
                    if (r != 0).any():  # Only accumulate non-zero rewards to handle sparse rewards
                        reward_func[s, a, s_prime] += np.sum(r) if isinstance(r, (list, np.ndarray)) else r
                        reward_func_count[s, a, s_prime] += 1
                elif r != 0:  # Handle scalar rewards
                    reward_func[s, a, s_prime] += r
                    reward_func_count[s, a, s_prime] += 1

    # Normalize reward and transition functions
    for s in range(num_values):
        for a in range(num_act):
            summed = sum(trans_func[s, a])
            if summed > 0:
                trans_func[s, a] /= summed
            else:
                trans_func[s, a] = np.ones(num_values) / num_values
            for s_prime in range(num_values):
                if reward_func_count[s, a, s_prime] > 0:
                    reward_func[s, a, s_prime] /= reward_func_count[s, a, s_prime]
                else:
                    reward_func[s, a, s_prime] = 0  # Maintain zero if no rewards were accumulated

    return reward_func, trans_func


def feature_selection(data_in, all_states, weights, cost, feat_to_select, num_feat_values,
                      num_act=2, discount_factor=0.85, verbose=True):
    """
    Selects state features based on Q-values using the G-algorithm.

    Args:
        data_in (list): List with samples of the form <s0, s1, a, r>.
        all_states: List of all possible states.
        weights (list): List of weights of the multiple rewards.
        cost: Cost of using human feedback.
        feat_to_select (list): Candidate features.
        num_feat_values: Number of features to select.
        num_act (int): Number of possible actions.
        discount_factor (float): Discount factor for Q-value computation.
        verbose (bool): Whether to print detailed output.

    Returns:
        list: Selected features.
        list: Criteria for selected features.
        list: New current state representations.
        list: New next state representations.
    """
    data = deepcopy(data_in)

    # Extract rewards and actions
    reward_list = list(np.array(data, dtype=object)[:, 3])
    list_of_actions = list(np.array(data, dtype=object)[:, 2])

    # Extract current and next state values
    list_of_s0 = list(np.array(data, dtype=object)[:, 0])
    list_of_s1 = list(np.array(data, dtype=object)[:, 1])

    # Assign weighted rewards to data
    for i in range(len(reward_list)):
        data[i][3] = reward_list[i]

    num_feat = len(feat_to_select)
    num_feat_to_select = len(num_feat_values)

    # Initialize Q-values and p-values for the first feature
    q_values = np.zeros((num_feat, num_feat_values[0], num_act))
    p_values = np.zeros(num_feat)
    num_values_first = num_feat_values[0]

    # Get state representations and calculate Q-values for each feature
    for f in range(num_feat):
        new_states, new_next_states = get_state_representation(all_states, list_of_s0, list_of_s1, num_values_first)

        # Debugging: Print the sizes of new_states, new_next_states, and data
        print(f"Length of data: {len(data)}")
        print(f"Length of new_states: {len(new_states)}")
        print(f"Length of new_next_states: {len(new_next_states)}")

        # Update the states in data
        for i in range(len(data)):
            data[i][0] = new_states[i]
            data[i][1] = new_next_states[i]

        # Convert lists to tuples before using them in a set
        unique_new_states = list(set(tuple(state) for state in new_states))
        unique_new_next_states = list(set(tuple(state) for state in new_next_states))

        # Approximate dynamics and reward function
        weighted_reward_func, trans_func = approx_dynamics_blocks(
            data, num_values_first, num_act, unique_new_states,
            unique_new_next_states, f, [])

        # Calculate Q-values
        q_values[f], _ = qval.get_Q_values_opt_policy(discount_factor, trans_func, weighted_reward_func)

        # Perform ANOVA to calculate p-values for the feature
        q_values_blocks = [q_values[f][b_ind] for b_ind in range(num_values_first)]
        p_values[f] = stats.f_oneway(*q_values_blocks)[1]

    # Debugging: Print p_values
    print(f"p_values: {p_values}")

    # Select the feature with the minimum p-value
    min_val = min(p_values)
    feat_min_p_val = [i for i in range(num_feat) if p_values[i] == min_val]
    feat_sel = [feat_to_select[feat_min_p_val[0]]] if len(feat_min_p_val) == 1 else [feat_to_select[random.choice(feat_min_p_val)]]
    feat_sel_criteria = ["Min p-value: " + str(round(min_val, 4))]

    if verbose:
        print("Feature selected:", feat_sel[-1])
        print("Criterion:", feat_sel_criteria[-1])

    # Iterate to select the remaining features
    for k in range(1, num_feat_to_select):
        feat_not_sel = [f for f in feat_to_select if f not in feat_sel]
        num_feat_not_sel = len(feat_not_sel)
        q_values_2 = np.zeros(num_feat_not_sel)
        p_values_2 = np.zeros(num_feat_not_sel)

        for f_ind, f in enumerate(feat_not_sel):
            num_values = num_feat_values[k]

            new_states, new_next_states = get_state_representation(all_states, list_of_s0, list_of_s1, num_values)

            for i in range(len(data)):
                data[i][0] = new_states[i]
                data[i][1] = new_next_states[i]

            # Approximate dynamics and reward function
            weighted_reward_func, trans_func = approx_dynamics_blocks(
                data, len(list(set(new_states))) + 1, num_act, list(set(new_states)),
                list(set(new_next_states)), f, feat_sel)

            # Calculate the Q-values
            q_values_2[f_ind], _ = qval.get_Q_values_opt_policy(discount_factor, trans_func, weighted_reward_func)

            # Calculate p-values using ANOVA
            q_values_blocks = [q_values_2[f_ind][b_ind] for b_ind in range(num_values)]
            p_values_2[f_ind] = stats.f_oneway(*q_values_blocks)[1]
            # Select the next feature with the minimum p-value
            min_val_curr = min(p_values_2)
            feat_min_p_val = [i for i in range(num_feat_not_sel) if p_values_2[i] == min_val_curr]

            if len(feat_min_p_val) == 1:
                feat_sel.append(feat_not_sel[feat_min_p_val[0]])
                feat_sel_criteria.append("Min p-value: " + str(round(min_val_curr, 4)))
            else:
                feat_sel.append(feat_not_sel[random.choice(feat_min_p_val)])
                feat_sel_criteria.append(
                    "Min p-value: " + str(round(min_val_curr, 4)) + " random from " + str(feat_min_p_val))

            if verbose:
                print("Feature selected:", feat_sel[-1])
                print("Criterion:", feat_sel_criteria[-1])

        # Extract the selected states in the correct representation format
        s0 = [list(np.take(state, feat_sel)) for state in np.array(data, dtype=object)[:, 0]]
        s1 = [list(np.take(state, feat_sel)) for state in np.array(data, dtype=object)[:, 1]]

        return feat_sel, feat_sel_criteria, s0, s1


def get_single_state_representation(all_states, s0, s1, num_values):
    """
    Converts a single state feature into a new format with a specified number of possible values.

    Args:
        all_states (list): List of all state values for the feature.
        s0 (list): The current state values for the feature.
        s1 (list): The next state values for the feature.
        num_values (int): The number of possible values for the feature.

    Returns:
        tuple: A tuple containing the new current state representation and the new next state representation.
    """

    # Check if all_states, s0, and s1 are empty or contain invalid data
    if not all_states or not s0 or not s1:
        print("Warning: One of the inputs is empty or invalid.")
        return [], []

    # Flatten and filter for numeric values
    all_states_flat = [item for sublist in all_states if isinstance(sublist, (int, float, list))
                       for item in (sublist if isinstance(sublist, list) else [sublist])
                       if isinstance(item, (int, float))]

    # Ensure that s0 and s1 are also numeric
    s0 = [x for x in s0 if isinstance(x, (int, float))]
    s1 = [x for x in s1 if isinstance(x, (int, float))]

    if len(all_states_flat) == 0:
        raise ValueError("No valid numeric data found in all_states.")

    if num_values == 3:
        first_tertile_cutoff = np.percentile(all_states_flat, 33.33)
        second_tertile_cutoff = np.percentile(all_states_flat, 66.67)
        new_state = [
            0 if x <= first_tertile_cutoff else 1 if first_tertile_cutoff < x <= second_tertile_cutoff else 2
            for x in s0
        ]
        new_next_state = [
            0 if x <= first_tertile_cutoff else 1 if first_tertile_cutoff < x <= second_tertile_cutoff else 2
            for x in s1
        ]

    elif num_values == 2:
        cutoff = np.percentile(all_states_flat, 50)
        new_state = [0 if x <= cutoff else 1 for x in s0]
        new_next_state = [0 if x <= cutoff else 1 for x in s1]

    else:
        raise ValueError("Unsupported number of values: {}".format(num_values))

    return new_state, new_next_state


def get_state_representation(all_states, s0, s1, num_values):
    """
    Converts the given state values into a new format with a specified number of possible values for each feature.

    Args:
        all_states (list): List of all state values for the features.
        s0 (list): The current state.
        s1 (list): The next state.
        num_values (int): The number of possible values for each feature.

    Returns:
        tuple: A tuple containing the new current state representation and the new next state representation.
    """

    new_states = []
    new_next_states = []

    # Check the shape and type of the inputs
    print(f"Shape of all_states: {len(all_states)} elements")
    print(f"Shape of s0: {len(s0)} elements")
    print(f"Shape of s1: {len(s1)} elements")

    # Explicitly check if each element of s0 is a list
    if not all(isinstance(item, list) for item in s0):
        print("Error: s0 is not a 2D list. Here's what s0 actually looks like:")
        for index, item in enumerate(s0[:5]):  # Print the first 5 elements for inspection
            print(f"Row {index}: Type - {type(item)}, Content - {item}")
        return [], []

    if not all(isinstance(item, list) for item in s1):
        print("Error: s1 is not a 2D list. Here's what s1 actually looks like:")
        for index, item in enumerate(s1[:5]):  # Print the first 5 elements for inspection
            print(f"Row {index}: Type - {type(item)}, Content - {item}")
        return [], []

    for i in range(len(s0[0])):  # For each feature
        try:
            list_of_s0 = list(np.array(s0, dtype=object)[:, i])
            list_of_s1 = list(np.array(s1, dtype=object)[:, i])
            list_of_all_states = list(np.array(all_states, dtype=object)[:, i])
        except IndexError as e:
            print(f"IndexError encountered: {e}")
            print(f"  s0 length: {len(s0)}, s1 length: {len(s1)}, all_states length: {len(all_states)}")
            return [], []

        if not list_of_s0 or not list_of_s1 or not list_of_all_states:
            print(f"Warning: Data missing for feature {i}.")
            continue

        # Get representation of current and next state for the single feature
        new_state, new_next_state = get_single_state_representation(
            list_of_all_states, list_of_s0, list_of_s1, num_values)

        if new_state and new_next_state:  # Only add if the states are valid
            new_states.append(new_state)
            new_next_states.append(new_next_state)

    if len(new_states) == 0 or len(new_next_states) == 0:
        print("Error: No valid state representations were generated.")
        return [], []

    # Transpose the lists to get them in the correct format
    new_states = list(map(list, zip(*new_states)))
    new_next_states = list(map(list, zip(*new_next_states)))

    print(f"New states generated: {len(new_states)}")
    print(f"New next states generated: {len(new_next_states)}")

    return new_states, new_next_states

