import numpy as np


def calculate_sparse_reward(row, option=1, no_identity_change_penalty=0.5):
    """
    Calculate the reward based on changes in smoking frequency and identity.
    Rewards are only calculated if the session number is greater than or equal to 5.

    Args:
        row (Series): A row of data containing the necessary fields for reward calculation.
        option (int): Option to decide which reward logic to apply.
        no_identity_change_penalty (float): Penalty for no identity change (used in option 1).

    Returns:
        tuple: (Frequency_Change_Reward, Identity_Change_Reward)
    """

    if row['session_num'] >= 5:  # Only apply rewards for session 5 and beyond
        if option == 1:
            # Frequency change reward calculation
            frequency_change = row['combined_PIV_Smoking_Freq_Vaping_Freq_x'] - row[
                'combined_Smoking_Freq_Vaping_Freq_y']
            max_frequency_change = 7  # Assuming the scale is 1 to 8
            frequency_change_reward = max(min(frequency_change / max_frequency_change, 1), -1)

            # Identity change reward calculation
            pre_identity_decrease_cols = ['Smok-vap_identity_1', 'Smok-vap_identity_2', 'Smok-vap_identity_3']
            post_identity_decrease_cols = ['combined_SSI_1_VSI_1', 'combined_SSI_2_VSI_2', 'combined_SSI_3_VSI_3']

            pre_identity_increase_cols = [
                'Quitter_SelfIdentity_1', 'Quitter_SelfIdentity_2', 'Quitter_SelfIdentity_4',
                'Not_smok-vap_identit_1', 'Not_smok-vap_identit_2', 'Not_smok-vap_identit_3'
            ]
            post_identity_increase_cols = [
                'combined_QSI_Smoke_1_QSI_Vape_1', 'combined_QSI_Smoke_2_QSI_Vape_2',
                'combined_QSI_Smoke_3_QSI_Vape_3', 'combined_NSSI_1_NVSI_1',
                'combined_NSSI_2_NVSI_2', 'combined_NSSI_3_NVSI_3'
            ]

            # Calculate identity changes
            identity_decrease_changes = [
                pre - post for pre, post in zip(row[pre_identity_decrease_cols], row[post_identity_decrease_cols])
            ]
            identity_increase_changes = [
                post - pre for pre, post in zip(row[pre_identity_increase_cols], row[post_identity_increase_cols])
            ]

            # Sum the identity change rewards
            total_identity_change = sum(identity_decrease_changes) + sum(identity_increase_changes)

            # Normalize identity change reward
            max_possible_identity_change = len(pre_identity_decrease_cols + pre_identity_increase_cols) * 5
            identity_change_reward = total_identity_change / max_possible_identity_change if max_possible_identity_change != 0 else 0
            identity_change_reward = max(min(identity_change_reward, 1), -1)

        elif option == 2:
            # Simple binary reward based on conditions
            frequency_change_reward = 1 if row['combined_Smoking_Freq_Vaping_Freq_y'] == 1 else -1
            identity_change_reward = 1 if sum([
                row['combined_SSI_1_VSI_1'], row['combined_SSI_2_VSI_2'], row['combined_SSI_3_VSI_3'],
                row['combined_QSI_Smoke_1_QSI_Vape_1'], row['combined_QSI_Smoke_2_QSI_Vape_2'],
                row['combined_QSI_Smoke_3_QSI_Vape_3'], row['combined_NSSI_1_NVSI_1'],
                row['combined_NSSI_2_NVSI_2'], row['combined_NSSI_3_NVSI_3']
            ]) > 0 else -1

    else:
        # If session number is less than 5, rewards are zero
        frequency_change_reward = 0
        identity_change_reward = 0

    return frequency_change_reward, identity_change_reward


def get_weighted_reward(rewards, weights, previous_human_support, cost_per_support, session_num):
    """
    Calculate weighted reward considering action cost.

    Args:
        rewards (list): List of rewards.
        weights (list): Weights for each reward.
        previous_human_support (int): Cumulative sum of previous 'human_support_after_session' values.
        cost_per_support (float): Cost associated with each support session.
        session_num (int): The current session number.

    Returns:
        float: Weighted reward.
    """
    if session_num >= 5:
        cost = previous_human_support * cost_per_support
        weighted_reward = np.dot(rewards, weights) - cost

        return weighted_reward
    else:
        return 0  # No reward should be applied before session 5


def get_weighted_rewards_list(df, weights, cost):
    """
    Generate a list of weighted rewards for each sample in the DataFrame, considering the action cost.

    Args:
        df (DataFrame): The input DataFrame containing rewards and actions.
        weights (list): Weights for each reward component.
        cost (float): Cost associated with the actions when 'human_support_after_session' is 1.

    Returns:
        list: List of weighted rewards.
    """
    rewards_list = []
    for index, row in df.iterrows():
        rewards = [row['Frequency_Change_Reward'], row['Identity_Change_Reward']]
        action = row['human_support_after_session']
        weighted_reward = get_weighted_reward(rewards, weights, action, cost)
        rewards_list.append(weighted_reward)
    return rewards_list
